#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <SDL/SDL_opengl.h>
#include <SDL/SDL_video.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include "render.h"
#include "image.h"

//
// Scales the input values down to a 
// range of 0-1
//
#define SCLFACTOR 0.004f

//
// A map from the 255 input values down to 50
// data points that uses a log scale, modified
// so as not to have too many repeated values
//
static int map50[] = {
0,
0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 
2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 
6, 7, 7, 8, 9, 10, 11, 12, 14, 15, 
17, 19, 21, 23, 26, 28, 31, 35, 38, 43, 
47, 52, 57, 63, 70, 77, 85, 94, 104, 115, 
126
};

//
// A map from the 255 input values down to 100
// data points that uses a log scale, modified
// so as not to have too many repeated values
//
static int map100[] = {
0,
0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 
2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 
5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 
7, 8, 8, 8, 8, 9, 9, 9, 9, 10,
10, 10, 10, 11, 11, 11, 12, 12, 12, 13,
13, 13, 14, 14, 14, 15, 15, 16, 16, 17, 
17, 18, 19, 20, 21, 22, 23, 24, 26, 27, 
28, 30, 31, 33, 35, 37, 38, 40, 43, 45, 
47, 49, 52, 55, 57, 60, 63, 67, 70, 74, 
77, 81, 85, 90, 94, 99, 104, 109, 115, 120,
126 };


//
// Definitions of the camera, lights, and various materials
//
static GLfloat camera_position[] = { 7.5f, 1.5f, 0.0f };
static GLfloat camera_up[]       = { 0.0f, 1.0f, 0.0f };
static GLfloat camera_at[]       = { 7.5f, 1.0f, 5.0f };

static GLfloat white_light[]   = { 1.0f, 1.0f, 1.0f, 1.0f };
static GLfloat light_position[]  = { 20.0f, 20.0f, -15.0f, 0.0f };

static GLfloat land_diffuse[] = { 0.65f, 0.46f, 0.14f };
static GLfloat land_ambient[] = { 0.1f, 0.1f, 0.1f };
static GLfloat land_specular[] = { 1.0f, 1.0f, 1.0f, 0.0f };
static GLfloat land_shininess[] = { 10.0f };

static GLfloat water_diffuse[] = { 0.0f, 0.2f, 0.8f };
static GLfloat water_ambient[] = { 0.1f, 0.15f, 0.3f };
static GLfloat water_specular[] = { 1.0f, 1.0f, 1.0f, 0.0f };
static GLfloat water_shininess[] = { 70.0f };

static GLfloat sky_diffuse[] = { 1.0f, 1.0f, 1.0f };
static GLfloat sky_ambient[] = { 0.7f, 0.7f, 0.7f };
static GLfloat sky_specular[] = { 1.0f, 1.0f, 1.0f, 0.0f };
static GLfloat sky_shininess[] = { 70.0f };


// Texture handle for the one water/sky texture we are using
static GLuint texture[1];

// Wheether it has allocated space or not
static int initialized = 0;

// Whether it is currently running or not
static int paused = 0;

// The shape of the terrain to use
static TerrainType type = CANYON;

// The surface used for output
static SDL_Surface *screen = NULL;

// The array of values for the heightfield
static GLfloat *data;

// Default screen sizes
static int screenwidth = 640;
static int screenheight = 480;

// Size of the data array
static int datawidth = 100;
static int datadepth = 60;

// The current "active" row in the heightfield where heights
// are being added from the spectrum data
static int hf_row = 0;

// Sky rotation
static GLfloat rot = -30.0f;


//
// Generates Mipmapped textures from the data in image.h
//
void GenerateTextures() {
  glGenTextures(1, texture);
  glBindTexture(GL_TEXTURE_2D, texture[0]);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB8, 200, 100, GL_RGB, 
		    GL_UNSIGNED_BYTE, (GLvoid*) &water_image);
}


//
// Sets up viewport when window is resized
//
void Reshape() {
  if (!initialized) return;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  glViewport(0, 0, screenwidth, screenheight);
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

  gluPerspective(35.0f, (GLfloat)screenwidth / (GLfloat)screenheight,
		 0.01f, 1000.0f);

  glMatrixMode(GL_MODELVIEW);
}


//
// Takes a row of spectrum data and maps it to values in
// the heightfield
//
void UpdateHeightField(gint16 *left, gint16 *right) {
  int i;
  int row_offset = datawidth * hf_row;
  int half = datawidth / 2;

  //correct for improper type
  guint16 *l = (guint16*) left;
  guint16 *r = (guint16*) right;
  
  if (!initialized) return;

  if (!paused) {
    switch (type) {
    case CANYON:
      for (i = 0; i < half; i++) {
	int idx = datawidth-i-1;
	data[row_offset + idx] = (sqrt(0.5f*(GLfloat)r[map50[i]] + 
				       (GLfloat)r[map50[i+1]] + 
				       0.5f*(GLfloat)r[map50[i+2]])) * SCLFACTOR;
	data[row_offset + i] = (sqrt(0.5f*(GLfloat)l[map50[i]] + 
				     (GLfloat)l[map50[i+1]] + 
				     0.5f*(GLfloat)l[map50[i+2]])) * SCLFACTOR;
      }
      break;

    case RIDGE:
      for (i = 0; i < half; i++) {
	int idx = half-i-1;
	data[row_offset + half + i] = (sqrt(0.5f*(GLfloat)r[map50[i]] + 
				       (GLfloat)r[map50[i+1]] + 
				       0.5f*(GLfloat)r[map50[i+2]])) * SCLFACTOR;
	data[row_offset + idx] = (sqrt(0.5f*(GLfloat)l[map50[i]] + 
				     (GLfloat)l[map50[i+1]] + 
				     0.5f*(GLfloat)l[map50[i+2]])) * SCLFACTOR;
      }
      break;

    case SLOPE:
      for (i = 0; i < datawidth; i++) {
	int idx = datawidth-i-1;
	data[row_offset + idx] = (sqrt(0.5f*(GLfloat)r[map100[i]] + 
				       (GLfloat)r[map100[i+1]] + 
				       0.5f*(GLfloat)r[map100[i+2]])) * SCLFACTOR;
      }
      break;
    }
  }
}

//
// Returns the normal at a given point in the heightfield
//
static inline void GetNormal(GLfloat *norm, int x, int y, int ny) {
  GLfloat dx = (data[y*datawidth + x + 1] - data[y*datawidth + x]) / 2.0f;
  GLfloat dy = (data[ny*datawidth + x] - data[y*datawidth + x]) / 2.0f;

  norm[1] = 1.0f;
  norm[0] = -dx * datawidth;
  norm[2] = -dy * datadepth;
}


//
// Draws a single frame (the height field, sea, and sky)
//
void DrawFrame() {
  int i,j;
  GLfloat norm[3];
  GLfloat offset;
  GLfloat stepx = 1.0f / (datawidth);
  GLfloat stepy = 1.0f / (datadepth);

  if (!initialized) return;
 
  //clear the screen
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  //bind the land material
  glDisable(GL_TEXTURE_2D);
  glMaterialfv(GL_FRONT, GL_AMBIENT, land_ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, land_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, land_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, land_shininess);
  
  //set up the lights and camera
  glLoadIdentity();
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  gluLookAt(camera_position[0], camera_position[1], camera_position[2],
	    camera_at[0], camera_at[1], camera_at[2],
	    camera_up[0], camera_up[1], camera_up[2]);

  //set up scaling for the land
  glPushMatrix();
  switch (type) {
  case RIDGE:
    glScalef(15.0f, 1.0f, 30.0f);
    break;
  case CANYON:
    glScalef(15.0f, 3.0f, 30.0f);
    break;
  case SLOPE:
    glTranslatef(-2.0f, 0.0f, 0.0f);
    glScalef(15.0f, 1.5f, 30.0f);
    break;
  }


  //draw the patch
  for (j = 1; j < datadepth - 2; j++) {
    int row = (j + hf_row + 1) % datadepth;
    int nrow = (row == datadepth-1) ? 0 : row + 1;
    int n2row = (nrow == datadepth-1) ? 0 : nrow + 1;

    //we use a separate triangle strip for each row
    glBegin(GL_TRIANGLE_STRIP);

    for (i = 1; i < datawidth - 1; i++) {
      GetNormal(norm, i, row, nrow);
      glNormal3f(norm[0], norm[1], norm[2]);
      glVertex3f(i*stepx, data[row*datawidth + i], j*stepy);

      GetNormal(norm, i, nrow, n2row);
      glNormal3f(norm[0], norm[1], norm[2]);
      glVertex3f(i*stepx, data[(nrow)*datawidth + i], (j+1)*stepy);
    }

    glEnd();
  }

  //bind the water material
  glBindTexture(GL_TEXTURE_2D, texture[0]);
  glMaterialfv(GL_FRONT, GL_AMBIENT, water_ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, water_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, water_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, water_shininess);

  //draw the water as a big square (using a texture map)
  offset = 2.0f * hf_row / datadepth;
  glEnable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
  glNormal3f(0.0f, 1.0f, 0.0f);
  glTexCoord2f(0.0f, offset + 0.0f);
  glVertex3f(-10.0f, 0.05f, -1.0f);

  glTexCoord2f(0.0f, offset + 4.0f);
  glVertex3f(-10.0f, 0.05f, 1.0f);

  glTexCoord2f(40.0f, offset + 4.0f);
  glVertex3f(10.0f, 0.05f, 1.0f);

  glTexCoord2f(40.0f, offset + 0.0f);
  glVertex3f(10.0f, 0.05f, -1.0f);

  glEnd();
  glPopMatrix();

  //bind the sky material
  glBindTexture(GL_TEXTURE_2D, texture[0]);
  glMaterialfv(GL_FRONT, GL_AMBIENT, sky_ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, sky_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, sky_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, sky_shininess);

  //draw the sky, also a big texture-mapped square
  glPushMatrix();
  glTranslatef(-15.0f, 7.0f, 30.0f);
  glRotatef(6.0f, 1.0f, 0.0f, 0.0f);
  glScalef(200.0f, 1.0f, 200.0f);
  glRotatef(rot, 0.0f, 1.0f, 0.0f);

  rot = rot + 0.1f;
  if (rot > 360.0f) rot -= 360.0f;

  glBegin(GL_QUADS);
  glNormal3f(0.0f, 1.0f, 0.0f);
  glTexCoord2f(0.0f, 8.0f);
  glVertex3f(-1.0f, 0.0f, 1.0f);

  glTexCoord2f(8.0f, 8.0f);
  glVertex3f(1.0f, 0.0f, 1.0f);

  glTexCoord2f(8.0f, 0.0f);
  glVertex3f(1.0f, 0.0f, -1.0f);

  glTexCoord2f(0.0f, 0.0f);
  glVertex3f(-1.0f, 0.0f, -1.0f);
  glEnd();
  glPopMatrix();

  glPopMatrix();
  glFlush();

  //update the current data row
  if (!paused) {
    hf_row = (hf_row + 1) % datadepth;
  }

  //flush to the screen
  SDL_GL_SwapBuffers();
}


//
// Checks for events, then calls DrawFrame()
//
void TerrainRender() {
  SDL_Event event;

  while (SDL_PollEvent(&event)) {
    if (event.type == SDL_KEYDOWN) {
      switch (event.key.keysym.sym) {
	//the "F" key toggles full screen mode
        case SDLK_f:
	  SDL_WM_ToggleFullScreen(screen);
	  break;

	//the "P" key pauses the terrain update (not the music)
        case SDLK_p:
	  paused = paused ? 0 : 1;
	  break;

	//the "M" key switches mode between the 3 types of terrain
        case SDLK_m:
	  if (type == CANYON) type = RIDGE;
	  else if (type == RIDGE) type = SLOPE;
	  else if (type == SLOPE) type = CANYON;
	  break;

        default:
	  break;
      }
    }
  }
  
  DrawFrame();
}


//
// Initializes the heightfield, SDL, and OpenGL stuff
//
void TerrainInit() {
  //allocate and zero the data array
  data = (GLfloat *) malloc(datawidth * datadepth * sizeof(GLfloat));
  if (data == NULL) {
    printf("Out of memory\n");
    return;
  }
  memset(data, 0, datawidth * datadepth * sizeof(GLfloat));

  //initialize SDL
  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_EVENTTHREAD)) {
    printf("Could not load SDL\n");
    return;
  }
  SDL_WM_SetCaption("Terrain Plugin", NULL);


  //set up SDL for OpenGL
  SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
  SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
  SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
 
  screen = SDL_SetVideoMode(screenwidth, screenheight, 16, SDL_OPENGL);
  if (screen == NULL) {
    printf("Could not set OpenGL video mode\n");
    return;
  }

  //set up OpenGL lighting
  glLightfv(GL_LIGHT0, GL_AMBIENT, white_light);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, white_light);

  //set up OpenGL options
  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_NORMALIZE);

  //create texture maps
  GenerateTextures();

  initialized = 1;
  Reshape();
}

//
// Frees allocated memory
//
void TerrainDestroy() {
  if (data != NULL) {
    free(data);
    data = NULL;
  }
  if (screen != NULL) {
    SDL_FreeSurface(screen);
    screen = NULL;
  }

  if (initialized) {
    glDeleteTextures(1, texture);
  }
  initialized = 0;

  SDL_Quit();
}
  
